#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""A bot client might go here... eventually..."""
__all__ = ["BotBase", "StatelessBot"]

import abc
import datetime
import logging
import typing

from hikari import events
from hikari.clients import configs
from hikari.clients import gateway_managers
from hikari.clients import rest_clients
from hikari.clients import runnable
from hikari.clients import shard_clients
from hikari.internal import more_asyncio
from hikari.internal import more_logging
from hikari.state import event_dispatchers
from hikari.state import event_managers
from hikari.state import stateless_event_managers


class BotBase(runnable.RunnableClient, event_dispatchers.EventDispatcher):
    """An abstract base class for a bot implementation.

    Parameters
    ----------
    config : :obj:`hikari.clients.configs.BotConfig`
        The config object to use.
    event_manager : ``hikari.state.event_managers.EventManager``
        The event manager to use.
    """

    #: The config for this bot.
    #:
    #: :type: :obj:`hikari.clients.configs.BotConfig`
    config: configs.BotConfig

    #: The event manager for this bot.
    #:
    #: :type: a subclass of :obj:`hikari.state.event_managers.EventManager`
    event_manager: event_managers.EventManager

    #: The gateway for this bot.
    #:
    #: :type: :obj:`hikari.clients.gateway_managers.GatewayManager` [ :obj:`hikari.clients.shard_clients.ShardClient` ]
    gateway: gateway_managers.GatewayManager[shard_clients.ShardClient]

    #: The logger to use for this bot.
    #:
    #: :type: :obj:`logging.Logger`
    logger: logging.Logger

    #: The REST HTTP client to use for this bot.
    #:
    #: :type: :obj:`hikari.clients.rest_clients.RESTClient`
    rest: rest_clients.RESTClient

    @abc.abstractmethod
    def __init__(self, config: configs.BotConfig, event_manager: event_managers.EventManager) -> None:
        super().__init__(more_logging.get_named_logger(self))
        self.config = config
        self.event_manager = event_manager
        self.gateway = NotImplemented
        self.rest = rest_clients.RESTClient(self.config)

    async def start(self):
        gateway_bot = await self.rest.fetch_gateway_bot()

        self.logger.info(
            "You have sent an IDENTIFY %s time(s) before now, and have %s remaining. This will reset at %s.",
            gateway_bot.session_start_limit.total - gateway_bot.session_start_limit.remaining,
            gateway_bot.session_start_limit.remaining,
            datetime.datetime.now() + gateway_bot.session_start_limit.reset_after,
        )

        shard_count = self.config.shard_count if self.config.shard_count else gateway_bot.shard_count
        shard_ids = self.config.shard_ids if self.config.shard_ids else [*range(shard_count)]

        self.gateway = gateway_managers.GatewayManager(
            config=self.config,
            url=gateway_bot.url,
            raw_event_consumer_impl=self.event_manager,
            shard_ids=shard_ids,
            shard_count=shard_count,
            dispatcher=self.event_manager.event_dispatcher.dispatch_event,
        )

        await self.gateway.start()

    async def close(self):
        await self.gateway.close()
        self.event_manager.event_dispatcher.close()
        await self.rest.close()

    async def join(self) -> None:
        await self.gateway.join()

    def add_listener(
        self, event_type: typing.Type[event_dispatchers.EventT], callback: event_dispatchers.EventCallbackT
    ) -> event_dispatchers.EventCallbackT:
        return self.event_manager.event_dispatcher.add_listener(event_type, callback)

    def remove_listener(
        self, event_type: typing.Type[event_dispatchers.EventT], callback: event_dispatchers.EventCallbackT
    ) -> event_dispatchers.EventCallbackT:
        return self.event_manager.event_dispatcher.remove_listener(event_type, callback)

    def wait_for(
        self,
        event_type: typing.Type[event_dispatchers.EventT],
        *,
        timeout: typing.Optional[float],
        predicate: event_dispatchers.PredicateT,
    ) -> more_asyncio.Future:
        return self.event_manager.event_dispatcher.wait_for(event_type, timeout=timeout, predicate=predicate)

    def dispatch_event(self, event: events.HikariEvent) -> more_asyncio.Future[typing.Any]:
        return self.event_manager.event_dispatcher.dispatch_event(event)


class StatelessBot(BotBase):
    """Bot client without any state internals.

    Parameters
    ----------
    config : :obj:`hikari.clients.configs.BotConfig`
        The config object to use.
    """

    def __init__(self, config: configs.BotConfig) -> None:
        super().__init__(config, stateless_event_managers.StatelessEventManagerImpl())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe message embeds on Discord."""
__all__ = [
    "Embed",
    "EmbedThumbnail",
    "EmbedVideo",
    "EmbedImage",
    "EmbedProvider",
    "EmbedAuthor",
    "EmbedFooter",
    "EmbedField",
]

import datetime
import typing

import attr

import hikari.internal.conversions
from hikari import colors
from hikari import entities
from hikari.internal import marshaller


@marshaller.marshallable()
@attr.s(slots=True)
class EmbedFooter(entities.HikariEntity, entities.Deserializable, entities.Serializable):
    """Represents a embed footer."""

    #: The footer text.
    #:
    #: :type: :obj:`str`
    text: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str)

    #: The URL of the footer icon.
    #:
    #: :type: :obj:`str`, optional
    icon_url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The proxied URL of the footer icon.
    #:
    #: :type: :obj:`str`, optional
    proxy_icon_url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)


@marshaller.marshallable()
@attr.s(slots=True)
class EmbedImage(entities.HikariEntity, entities.Deserializable, entities.Serializable):
    """Represents a embed image."""

    #: The URL of the image.
    #:
    #: :type: :obj:`str`, optional
    url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The proxied URL of the image.
    #:
    #: :type: :obj:`str`, optional
    proxy_url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The height of the image.
    #:
    #: :type: :obj:`int`, optional
    height: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=int, if_undefined=None)

    #: The width of the image.
    #:
    #: :type: :obj:`int`, optional
    width: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=int, if_undefined=None)


@marshaller.marshallable()
@attr.s(slots=True)
class EmbedThumbnail(entities.HikariEntity, entities.Deserializable, entities.Serializable):
    """Represents a embed thumbnail."""

    #: The URL of the thumbnail.
    #:
    #: :type: :obj:`str`, optional
    url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The proxied URL of the thumbnail.
    #:
    #: :type: :obj:`str`, optional
    proxy_url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The height of the thumbnail.
    #:
    #: :type: :obj:`int`, optional
    height: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=int, if_undefined=None)

    #: The width of the thumbnail.
    #:
    #: :type: :obj:`int`, optional
    width: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=int, if_undefined=None)


@marshaller.marshallable()
@attr.s(slots=True)
class EmbedVideo(entities.HikariEntity, entities.Deserializable, entities.Serializable):
    """Represents a embed video."""

    #: The URL of the video.
    #:
    #: :type: :obj:`str`, optional
    url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The height of the video.
    #:
    #: :type: :obj:`int`, optional
    height: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=int, if_undefined=None)

    #: The width of the video.
    #:
    #: :type: :obj:`int`, optional
    width: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=int, if_undefined=None)


@marshaller.marshallable()
@attr.s(slots=True)
class EmbedProvider(entities.HikariEntity, entities.Deserializable, entities.Serializable):
    """Represents a embed provider."""

    #: The name of the provider.
    #:
    #: :type: :obj:`str`, optional
    name: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The URL of the provider.
    #:
    #: :type: :obj:`str`, optional
    url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None, if_none=None)


@marshaller.marshallable()
@attr.s(slots=True)
class EmbedAuthor(entities.HikariEntity, entities.Deserializable, entities.Serializable):
    """Represents a embed author."""

    #: The name of the author.
    #:
    #: :type: :obj:`str`, optional
    name: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The URL of the author.
    #:
    #: :type: :obj:`str`, optional
    url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The URL of the author icon.
    #:
    #: :type: :obj:`str`, optional
    icon_url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The proxied URL of the author icon.
    #:
    #: :type: :obj:`str`, optional
    proxy_icon_url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)


@marshaller.marshallable()
@attr.s(slots=True)
class EmbedField(entities.HikariEntity, entities.Deserializable, entities.Serializable):
    """Represents a field in a embed."""

    #: The name of the field.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str, serializer=str)

    #: The value of the field.
    #:
    #: :type: :obj:`str`
    value: str = marshaller.attrib(deserializer=str, serializer=str)

    #: Whether the field should display inline. Defaults to ``False``.
    #:
    #: :type: :obj:`bool`
    is_inline: bool = marshaller.attrib(raw_name="inline", deserializer=bool, serializer=bool, if_undefined=False)


@marshaller.marshallable()
@attr.s(slots=True)
class Embed(entities.HikariEntity, entities.Deserializable, entities.Serializable):
    """Represents a embed."""

    #: The title of the embed.
    #:
    #: :type: :obj:`str`, optional
    title: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The description of the embed.
    #:
    #: :type: :obj:`str`, optional
    description: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The URL of the embed.
    #:
    #: :type: :obj:`str`, optional
    url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None)

    #: The timestamp of the embed.
    #:
    #: :type: :obj:`datetime.datetime`, optional
    timestamp: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=hikari.internal.conversions.parse_iso_8601_ts,
        serializer=lambda timestamp: timestamp.replace(tzinfo=datetime.timezone.utc).isoformat(),
        if_undefined=None,
    )

    color: typing.Optional[colors.Color] = marshaller.attrib(
        deserializer=colors.Color, serializer=int, if_undefined=None
    )

    #: The footer of the embed.
    #:
    #: :type: :obj:`EmbedFooter`, optional
    footer: typing.Optional[EmbedFooter] = marshaller.attrib(
        deserializer=EmbedFooter.deserialize, serializer=EmbedFooter.serialize, if_undefined=None
    )

    #: The image of the embed.
    #:
    #: :type: :obj:`EmbedImage`, optional
    image: typing.Optional[EmbedImage] = marshaller.attrib(
        deserializer=EmbedImage.deserialize, serializer=EmbedImage.serialize, if_undefined=None
    )

    #: The thumbnail of the embed.
    #:
    #: :type: :obj:`EmbedThumbnail`, optional
    thumbnail: typing.Optional[EmbedThumbnail] = marshaller.attrib(
        deserializer=EmbedThumbnail.deserialize, serializer=EmbedThumbnail.serialize, if_undefined=None
    )

    #: The video of the embed.
    #:
    #: :type: :obj:`EmbedVideo`, optional
    video: typing.Optional[EmbedVideo] = marshaller.attrib(
        deserializer=EmbedVideo.deserialize, serializer=EmbedVideo.serialize, if_undefined=None
    )

    #: The provider of the embed.
    #:
    #: :type: :obj:`EmbedProvider`, optional
    provider: typing.Optional[EmbedProvider] = marshaller.attrib(
        deserializer=EmbedProvider.deserialize, serializer=EmbedProvider.serialize, if_undefined=None
    )

    #: The author of the embed.
    #:
    #: :type: :obj:`EmbedAuthor`, optional
    author: typing.Optional[EmbedAuthor] = marshaller.attrib(
        deserializer=EmbedAuthor.deserialize, serializer=EmbedAuthor.serialize, if_undefined=None
    )

    #: The fields of the embed.
    #:
    #: :type: :obj:`typing.Sequence` [ :obj:`EmbedField` ], optional
    fields: typing.Optional[typing.Sequence[EmbedField]] = marshaller.attrib(
        deserializer=lambda fields: [EmbedField.deserialize(f) for f in fields],
        serializer=lambda fields: [f.serialize() for f in fields],
        if_undefined=None,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Discord-specific URIs that have to be hard-coded.

|internal|
"""

__all__ = [
    "generate_cdn_url",
]

import typing
import urllib.parse


#: The URL for the CDN.
#:
#: :type: :obj:`str`
BASE_CDN_URL: typing.Final[str] = "https://cdn.discordapp.com"

#: The URL for the REST API. This contains a version number parameter that
#: should be interpolated.
#:
#: :type: :obj:`str`
REST_API_URL: typing.Final[str] = "https://discordapp.com/api/v{0.version}"


def generate_cdn_url(*route_parts: str, fmt: str, size: typing.Optional[int]) -> str:
    """Generate a link for a Discord CDN media resource.

    Parameters
    ----------
    route_parts : :obj:`str`
        The string route parts that will be used to form the link.
    fmt : :obj:`str`
        The format to use for the wanted cdn entity, will usually be one of
        ``webp``, ``png``, ``jpeg``, ``jpg`` or ``gif`` (which will be invalid
        if the target entity doesn't have an animated version available).
    size : :obj:`int`, optional
        The size to specify for the image in the query string if applicable,
        should be passed through as ``None`` to avoid the param being set.

    Returns
    -------
    :obj:`str`
        The URL to the resource on the Discord CDN.
    """
    path = "/".join(urllib.parse.unquote(part) for part in route_parts)
    url = urllib.parse.urljoin(BASE_CDN_URL, "/" + path) + "." + str(fmt)
    query = urllib.parse.urlencode({"size": size}) if size is not None else None
    return f"{url}?{query}" if query else url



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""May contain nuts."""
__all__ = ["RESTClient"]

import datetime

from hikari import gateway_entities
from hikari.clients import configs


class RESTClient:
    """Stuff will go here when this is implemented..."""

    # TODO: FasterSpeeding: update this.
    def __init__(self, _: configs.RESTConfig) -> None:
        ...

    async def close(self):
        ...

    async def fetch_gateway_bot(self) -> gateway_entities.GatewayBot:
        # Stubbed placeholder.
        # TODO: replace with actual implementation.
        return gateway_entities.GatewayBot(
            url="wss://gateway.discord.gg",
            shard_count=1,
            session_start_limit=gateway_entities.SessionStartLimit(
                total=1000, remaining=999, reset_after=datetime.timedelta(days=0.9)
            ),
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Event management for stateless bots."""

__all__ = ["StatelessEventManagerImpl"]

from hikari import events
from hikari.state import event_dispatchers
from hikari.state import event_managers


class StatelessEventManagerImpl(event_managers.EventManager[event_dispatchers.EventDispatcher]):
    """Stateless event manager implementation for stateless bots.

    This is an implementation that does not rely on querying prior information to
    operate. The implementation details of this are much simpler than a stateful
    version, and are not immediately affected by the use of intents.
    """

    @event_managers.raw_event_mapper("CONNECTED")
    def on_connect(self, shard, _):
        """Handle CONNECTED events."""
        event = events.ConnectedEvent(shard=shard)
        self.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("DISCONNECTED")
    def on_disconnect(self, shard, _):
        """Handle DISCONNECTED events."""
        event = events.DisconnectedEvent(shard=shard)
        self.event_dispatcher.dispatch_event(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Asyncio extensions and utilities.

|internal|
"""
from __future__ import annotations

__all__ = ["Future", "Task", "completed_future"]

import asyncio
import contextvars
import typing

T = typing.TypeVar("T")
T_co = typing.TypeVar("T_co")

try:
    raise Exception
except Exception as ex:  # pylint:disable=broad-except
    tb = ex.__traceback__
    StackFrameType = type(tb.tb_frame)


# pylint:disable=unused-variable
@typing.runtime_checkable
class Future(typing.Protocol[T]):
    """Typed protocol representation of an :obj:`asyncio.Future`.

    You should consult the documentation for :obj:`asyncio.Future` for usage.
    """

    def result(self) -> T:
        """See :meth:`asyncio.Future.result`."""

    def set_result(self, result: T, /) -> None:
        """See :meth:`asyncio.Future.set_result`."""

    def set_exception(self, exception: Exception, /) -> None:
        """See :meth:`asyncio.Future.set_exception`."""

    def done(self) -> bool:
        """See :meth:`asyncio.Future.done`."""

    def cancelled(self) -> bool:
        """See :meth:`asyncio.Future.cancelled`."""

    def add_done_callback(
        self, callback: typing.Callable[[Future[T]], None], /, *, context: typing.Optional[contextvars.Context],
    ) -> None:
        """See :meth:`asyncio.Future.add_done_callback`."""

    def remove_done_callback(self, callback: typing.Callable[[Future[T]], None], /) -> None:
        """See :meth:`asyncio.Future.remove_done_callback`."""

    def cancel(self) -> bool:
        """See :meth:`asyncio.Future.cancel`."""

    def exception(self) -> typing.Optional[Exception]:
        """See :meth:`asyncio.Future.exception`."""

    def get_loop(self) -> asyncio.AbstractEventLoop:
        """See :meth:`asyncio.Future.get_loop`."""

    def __await__(self) -> typing.Coroutine[None, None, T]:
        ...


# pylint:enable=unused-variable


# pylint:disable=unused-variable
class Task(Future[T]):
    """Typed protocol representation of an :obj:`asyncio.Task`.

    You should consult the documentation for :obj:`asyncio.Task` for usage.
    """

    def get_stack(self, *, limit: typing.Optional[int] = None) -> typing.Sequence[StackFrameType]:
        """See :meth:`asyncio.Task.get_stack`."""

    def print_stack(self, *, limit: typing.Optional[int] = None, file: typing.Optional[typing.IO] = None) -> None:
        """See :meth:`asyncio.Task.print_stack`."""

    def get_name(self) -> str:
        """See :meth:`asyncio.Task.get_name`."""

    def set_name(self, value: str, /) -> None:
        """See :meth:`asyncio.Task.set_name`."""


# pylint:enable=unused-variable


@typing.overload
def completed_future() -> Future[None]:
    """Return a completed future with no result."""


@typing.overload
def completed_future(result: T, /) -> Future[T]:
    """Return a completed future with the given value as the result."""


def completed_future(result=None, /):
    """Create a future on the current running loop that is completed, then return it.

    Parameters
    ----------
    result : :obj:`typing.Any`
        The value to set for the result of the future.

    Returns
    -------
    :obj:`asyncio.Future`
        The completed future.
    """
    future = asyncio.get_event_loop().create_future()
    future.set_result(result)
    return future



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Hikari's models framework for writing Discord bots in Python."""
from hikari import audit_logs
from hikari import channels
from hikari import clients
from hikari import colors
from hikari import colours
from hikari import embeds
from hikari import emojis
from hikari import entities
from hikari import errors
from hikari import events
from hikari import gateway_entities
from hikari import guilds
from hikari import invites
from hikari import messages
from hikari import net
from hikari import oauth2
from hikari import permissions
from hikari import snowflakes
from hikari import state
from hikari import users
from hikari import voices
from hikari import webhooks
from hikari._about import __author__
from hikari._about import __copyright__
from hikari._about import __email__
from hikari._about import __license__
from hikari._about import __url__
from hikari._about import __version__
from hikari.audit_logs import *
from hikari.channels import *
from hikari.clients import *
from hikari.colors import *
from hikari.colours import *
from hikari.embeds import *
from hikari.emojis import *
from hikari.entities import *
from hikari.events import *
from hikari.gateway_entities import *
from hikari.guilds import *
from hikari.invites import *
from hikari.messages import *
from hikari.net import *
from hikari.oauth2 import *
from hikari.permissions import *
from hikari.snowflakes import *
from hikari.state import *
from hikari.users import *
from hikari.voices import *
from hikari.webhooks import *

# Import everything into this namespace.

__all__ = [
    *audit_logs.__all__,
    *channels.__all__,
    *clients.__all__,
    *colors.__all__,
    *colours.__all__,
    *embeds.__all__,
    *emojis.__all__,
    *entities.__all__,
    *events.__all__,
    *gateway_entities.__all__,
    *guilds.__all__,
    *invites.__all__,
    *messages.__all__,
    *net.__all__,
    *oauth2.__all__,
    *permissions.__all__,
    *snowflakes.__all__,
    *state.__all__,
    *users.__all__,
    *voices.__all__,
    *webhooks.__all__,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""API version enumeration."""
__all__ = ["HTTPAPIVersion", "GatewayVersion"]

import enum


class HTTPAPIVersion(enum.IntEnum):
    """Supported versions for the REST API."""

    #: The V6 API. This is currently the stable release that should be used unless you have a reason
    #: to use V7 otherwise.
    V6 = 6

    #: Development API version. This is not documented at the time of writing, and is subject to
    #: change at any time without warning.
    V7 = 7

    #: The recommended stable API release to default to.
    STABLE = V6


class GatewayVersion(enum.IntEnum):
    """Supported versions for the Gateway."""

    #: The V6 API. This is currently the stable release that should be used unless you have a reason
    #: to use V7 otherwise.
    V6 = 6

    #: Development API version. This is not documented at the time of writing, and is subject to
    #: change at any time without warning.
    V7 = 7

    #: The recommended stable API release to default to.
    STABLE = V6



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Anonymous system information that we have to provide to Discord when using their API.

This information contains details such as the version of Python you are using, and
the version of this library, the OS you are making requests from, etc.

This information is provided to enable Discord to detect that you are using a
valid bot and not attempting to abuse the API.
"""
__all__ = ["UserAgent"]

import typing

from hikari.internal import meta


class UserAgent(metaclass=meta.SingletonMeta):
    """Platform version info.

    Notes
    -----
    This is a singleton.
    """

    #: The version of the library.
    #:
    #: Example
    #: -------
    #: ``"hikari 1.0.1"``
    #:
    #: :type: :obj:`typing.Final` [ :obj:`str` ]
    library_version: typing.Final[str]

    #: The platform version.
    #:
    #: Example
    #: -------
    #: ``"CPython 3.8.2 GCC 9.2.0"``
    #:
    #: :type: :obj:`typing.Final` [ :obj:`str` ]
    platform_version: typing.Final[str]

    #: The operating system type.
    #:
    #: Example
    #: -------
    #: ``"Linux-5.4.15-2-MANJARO-x86_64-with-glibc2.2.5"``
    #:
    #: :type: :obj:`typing.Final` [ :obj:`str` ]
    system_type: typing.Final[str]

    #: The Hikari-specific user-agent to use in HTTP connections to Discord.
    #:
    #: Example
    #: -------
    #: ``"DiscordBot (https://gitlab.com/nekokatt/hikari; 1.0.1; Nekokatt) CPython 3.8.2 GCC 9.2.0 Linux"``
    #:
    #: :type: :obj:`typing.Final` [ :obj:`str` ]
    user_agent: typing.Final[str]

    def __init__(self):
        from hikari._about import __author__, __url__, __version__
        from platform import python_implementation, python_version, python_branch, python_compiler, platform

        self.library_version = f"hikari {__version__}"
        self.platform_version = self._join_strip(
            python_implementation(), python_version(), python_branch(), python_compiler()
        )
        self.system_type = platform()
        self.user_agent = f"DiscordBot ({__url__}; {__version__}; {__author__}) {python_version()} {self.system_type}"

        def __attr__(_):
            raise TypeError("cannot change attributes once set")

        self.__delattr__ = __attr__
        self.__setattr__ = __attr__

    @staticmethod
    def _join_strip(*args):
        return " ".join((arg.strip() for arg in args if arg.strip()))

    # Ignore docstring not starting in an imperative mood
    @property
    def websocket_triplet(self) -> typing.Dict[str, str]:  # noqa: D401
        """A dict representing device and library info.

        Returns
        -------
        :obj:`typing.Dict` [ :obj:`str`, :obj:`str` ]
            The object to send to Discord representing device info when
            IDENTIFYing with the gateway.
        """
        return {
            "$os": self.system_type,
            "$browser": self.library_version,
            "$device": self.platform_version,
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Provides the valid routes that can be used on the API, as well as mechanisms to aid with rate limit bucketing."""
__all__ = ["CompiledRoute", "RouteTemplate"]

import typing

DEFAULT_MAJOR_PARAMS = {"channel_id", "guild_id", "webhook_id"}


class CompiledRoute:
    """A compiled representation of a route ready to be made into a full e and to be used for a request.

    Parameters
    ----------
    method : :obj:`str`
        The HTTP method to use.
    path : :obj:`str`
        The path with any major parameters interpolated in.
    major_params_hash : :obj:`str`
        The part of the hash identifier to use for the compiled set of major parameters.
    """

    __slots__ = ("method", "major_params_hash", "compiled_path", "hash_code", "__weakref__")

    #: The method to use on the route.
    #:
    #: :type: :obj:`str`
    method: typing.Final[str]

    #: The major parameters in a bucket hash-compatible representation.
    #:
    #: :type: :obj:`str`
    major_params_hash: typing.Final[str]

    #: The compiled route path to use
    #:
    #: :type: :obj:`str`
    compiled_path: typing.Final[str]

    #: The hash code
    #:
    #: :type: :obj:`int`
    hash_code: typing.Final[int]

    def __init__(self, method: str, path_template: str, path: str, major_params_hash: str) -> None:
        self.method = method
        self.major_params_hash = major_params_hash
        self.compiled_path = path
        self.hash_code = hash((path_template, major_params_hash))

    def create_url(self, base_url: str) -> str:
        """Create the full URL with which you can make a request.

        Parameters
        ----------
        base_url : :obj:`str`
            The base of the URL to prepend to the compiled path.

        Returns
        -------
        :obj:`str`
            The full URL for the route.
        """
        return base_url + self.compiled_path

    def create_real_bucket_hash(self, initial_bucket_hash: str) -> str:
        """Create a full bucket hash from a given initial hash.

        The result of this hash will be decided by the value of the major
        parameters passed to the route during the compilation phase.

        Parameters
        ----------
        initial_bucket_hash: :obj:`str`
            The initial bucket hash provided by Discord in the HTTP headers
            for a given response.

        Returns
        -------
        :obj:`str`
            The input hash amalgamated with a hash code produced by the
            major parameters in this compiled route instance.
        """
        return initial_bucket_hash + ";" + self.major_params_hash

    def __hash__(self) -> int:
        return self.hash_code

    def __eq__(self, other) -> bool:
        return hash(self) == hash(other)

    def __repr__(self) -> str:
        this_type = type(self).__name__
        major_params = ", ".join(
            (
                f"method={self.method!r}",
                f"compiled_path={self.compiled_path!r}",
                f"major_params_hash={self.major_params_hash!r}",
            )
        )
        return f"{this_type}({major_params})"

    def __str__(self) -> str:
        return f"{self.method} {self.compiled_path}"


class RouteTemplate:
    """A template used to create compiled routes for specific parameters.

    These compiled routes are used to identify rate limit buckets.

    Parameters
    ----------
    path_template : :obj:`str`
        The template string for the path to use.
    major_params : :obj:`str`
        A collection of major parameter names that appear in the template path.
        If not specified, the default major parameter names are extracted and
        used in-place.
    """

    __slots__ = ("path_template", "major_params")

    #: The template string used for the path.
    #:
    #: :type: :obj:`str`
    path_template: typing.Final[str]

    #: Major parameter names that appear in the template path.
    #:
    #: :type: :obj:`typing.FrozenSet` [ :obj:`str` ]
    major_params: typing.Final[typing.FrozenSet[str]]

    def __init__(self, path_template: str, major_params: typing.Collection[str] = None) -> None:
        self.path_template = path_template
        if major_params is None:
            self.major_params = frozenset(p for p in DEFAULT_MAJOR_PARAMS if f"{{{p}}}" in path_template)
        else:
            self.major_params = frozenset(major_params)

    def compile(self, method: str, /, **kwargs: typing.Any) -> CompiledRoute:
        """Generate a formatted :obj:`CompiledRoute` for this route template.

        This takes into account any URL parameters that have been passed, and extracting
        the :attr:major_params" for bucket hash operations accordingly.

        Parameters
        ----------
        method : :obj:`str`
            The method to use.
        **kwargs : :obj:`typing.Any`
            Any parameters to interpolate into the route path.

        Returns
        -------
        :obj:`CompiledRoute`
            The compiled route.
        """
        major_hash_part = "-".join((str(kwargs[p]) for p in self.major_params))

        return CompiledRoute(method, self.path_template, self.path_template.format_map(kwargs), major_hash_part)

    def __repr__(self) -> str:
        this_type = type(self).__name__
        major_params = ", ".join((f"path_template={self.path_template!r}", f"major_params={self.major_params!r}",))
        return f"{this_type}({major_params})"

    def __str__(self) -> str:
        return self.path_template


_RT = typing.Final[RouteTemplate]

# Channels
CHANNEL: _RT = RouteTemplate("/channels/{channel_id}")
CHANNEL_DM_RECIPIENTS: _RT = RouteTemplate("/channels/{channel_id}/recipients/{user_id}")
CHANNEL_INVITES: _RT = RouteTemplate("/channels/{channel_id}/invites")
CHANNEL_MESSAGE: _RT = RouteTemplate("/channels/{channel_id}/messages/{message_id}")
CHANNEL_MESSAGES: _RT = RouteTemplate("/channels/{channel_id}/messages")
CHANNEL_MESSAGES_BULK_DELETE: _RT = RouteTemplate("/channels/{channel_id}/messages")
CHANNEL_PERMISSIONS: _RT = RouteTemplate("/channels/{channel_id}/permissions/{overwrite_id}")
CHANNEL_PIN: _RT = RouteTemplate("/channels/{channel_id}/pins/{message_id}")
CHANNEL_PINS: _RT = RouteTemplate("/channels/{channel_id}/pins")
CHANNEL_TYPING: _RT = RouteTemplate("/channels/{channel_id}/typing")
CHANNEL_WEBHOOKS: _RT = RouteTemplate("/channels/{channel_id}/webhooks")

# Reactions
ALL_REACTIONS: _RT = RouteTemplate("/channels/{channel_id}/messages/{message_id}/reactions")
REACTION_EMOJI: _RT = RouteTemplate("/channels/{channel_id}/messages/{message_id}/reactions/{emoji}")
REACTION_EMOJI_USER: _RT = RouteTemplate("/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/{used_id}")
REACTIONS: _RT = RouteTemplate("/channels/{channel_id}/messages/{message_id}/reactions/{emoji}")

# Guilds
GUILD: _RT = RouteTemplate("/guilds/{guild_id}")
GUILDS: _RT = RouteTemplate("/guilds")
GUILD_AUDIT_LOGS: _RT = RouteTemplate("/guilds/{guild_id}/audit-logs")
GUILD_BAN: _RT = RouteTemplate("/guilds/{guild_id}/bans/{user_id}")
GUILD_BANS: _RT = RouteTemplate("/guilds/{guild_id}/bans")
GUILD_CHANNELS: _RT = RouteTemplate("/guilds/{guild_id}/channels")
GUILD_EMBED: _RT = RouteTemplate("/guilds/{guild_id}/embed")
GUILD_EMOJI: _RT = RouteTemplate("/guilds/{guild_id}/emojis/{emoji_id}")
GUILD_EMOJIS: _RT = RouteTemplate("/guilds/{guild_id}/emojis")
GUILD_INTEGRATION: _RT = RouteTemplate("/guilds/{guild_id}/integrations/{integration_id}")
GUILD_INTEGRATIONS: _RT = RouteTemplate("/guilds/{guild_id}/integrations")
GUILD_INTEGRATION_SYNC: _RT = RouteTemplate("/guilds/{guild_id}/integrations/{integration_id}")
GUILD_INVITES: _RT = RouteTemplate("/guilds/{guild_id}/invites")
GUILD_MEMBERS: _RT = RouteTemplate("/guilds/{guild_id}/members")
GUILD_MEMBER: _RT = RouteTemplate("/guilds/{guild_id}/members/{user_id}")
GUILD_MEMBER_ROLE: _RT = RouteTemplate("/guilds/{guild_id}/members/{user_id}/roles/{role_id}")
GUILD_PRUNE: _RT = RouteTemplate("/guilds/{guild_id}/prune")
GUILD_ROLE: _RT = RouteTemplate("/guilds/{guild_id}/roles/{role_id}")
GUILD_ROLES: _RT = RouteTemplate("/guilds/{guild_id}/roles")
GUILD_VANITY_URL: _RT = RouteTemplate("/guilds/{guild_id}/vanity-url")
GUILD_VOICE_REGIONS: _RT = RouteTemplate("/guilds/{guild_id}/regions")
GUILD_WIDGET_IMAGE: _RT = RouteTemplate("/guilds/{guild_id}/widget.png")
GUILD_WEBHOOKS: _RT = RouteTemplate("/guilds/{guild_id}/webhooks")

# Invites
INVITE: _RT = RouteTemplate("/invites/{invite_code}")

# Users
USER: _RT = RouteTemplate("/users/{user_id}")

# @me
LEAVE_GUILD: _RT = RouteTemplate("/users/@me/guilds/{guild_id}")
OWN_CONNECTIONS: _RT = RouteTemplate("/users/@me/connections")  # OAuth2 only
OWN_DMS: _RT = RouteTemplate("/users/@me/channels")
OWN_GUILDS: _RT = RouteTemplate("/users/@me/guilds")
OWN_GUILD_NICKNAME: _RT = RouteTemplate("/guilds/{guild_id}/members/@me/nick")
OWN_USER: _RT = RouteTemplate("/users/@me")
OWN_REACTION: _RT = RouteTemplate("/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me")

# Voice
VOICE_REGIONS: _RT = RouteTemplate("/voice/regions")

# Webhooks
WEBHOOK: _RT = RouteTemplate("/webhooks/{webhook_id}")
WEBHOOK_WITH_TOKEN: _RT = RouteTemplate("/webhooks/{webhook_id}/{webhook_token}")
WEBHOOK_WITH_TOKEN_GITHUB: _RT = RouteTemplate("/webhooks/{webhook_id}/{webhook_token}/github")
WEBHOOK_WITH_TOKEN_SLACK: _RT = RouteTemplate("/webhooks/{webhook_id}/{webhook_token}/slack")

# OAuth2 API
OAUTH2_APPLICATIONS: _RT = RouteTemplate("/oauth2/applications")
OAUTH2_APPLICATIONS_ME: _RT = RouteTemplate("/oauth2/applications/@me")
OAUTH2_AUTHORIZE: _RT = RouteTemplate("/oauth2/authorize")
OAUTH2_TOKEN: _RT = RouteTemplate("/oauth2/token")
OAUTH2_TOKEN_REVOKE: _RT = RouteTemplate("/oauth2/token/revoke")

# Gateway
GATEWAY: _RT = RouteTemplate("/gateway")
GATEWAY_BOT: _RT = RouteTemplate("/gateway/bot")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Bitfield of permissions."""
__all__ = ["Permission"]

import enum


class Permission(enum.IntFlag):
    """Represents the permissions available in a given channel or guild.

    This is an int-flag enum. This means that you can **combine multiple
    permissions together** into one value using the bitwise-OR operator (``|``).

    .. code-block:: python

        my_perms = Permission.MANAGE_CHANNELS | Permission.MANAGE_GUILD

        your_perms = (
            Permission.CREATE_INSTANT_INVITE
            | Permission.KICK_MEMBERS
            | Permission.BAN_MEMBERS
            | Permission.MANAGE_GUILD
        )

    You can **check if a permission is present** in a set of combined
    permissions by using the bitwise-AND operator (``&``). This will return
    the int-value of the permission if it is present, or ``0`` if not present.

    .. code-block:: python

        my_perms = Permission.MANAGE_CHANNELS | Permission.MANAGE_GUILD

        if my_perms & Permission.MANAGE_CHANNELS:
            if my_perms & Permission.MANAGE_GUILD:
                print("I have the permission to both manage the guild and the channels in it!")
            else:
                print("I have the permission to manage channels!")
        else:
            print("I don't have the permission to manage channels!")

        # Or you could simplify it:

        if my_perms & (Permission.MANAGE_CHANNELS | Permission.MANAGE_GUILD):
            print("I have the permission to both manage the guild and the channels in it!")
        elif my_perms & Permission.MANAGE_CHANNELS:
            print("I have the permission to manage channels!")
        else:
            print("I don't have the permission to manage channels!")

    If you need to **check that a permission is not present**, you can use the
    bitwise-XOR operator (``^``) to check. If the permission is not present, it
    will return a non-zero value, otherwise if it is present, it will return ``0``.

    .. code-block:: python

        my_perms = Permission.MANAGE_CHANNELS | Permission.MANAGE_GUILD

        if my_perms ^ Permission.MANAGE_CHANNELS:
            print("Please give me the MANAGE_CHANNELS permission!")

    Lastly, if you need all the permissions set except the permission you want,
    you can use the inversion operator (``~``) to do that.

    .. code-block:: python

        # All permissions except ADMINISTRATOR.
        my_perms = ~Permission.ADMINISTRATOR

    """

    #: Empty permission.
    NONE = 0x0
    #: Allows creation of instant invites.
    CREATE_INSTANT_INVITE = 0x1
    #: Allows kicking members
    KICK_MEMBERS = 0x2
    #: Allows banning members.
    BAN_MEMBERS = 0x4
    #: Allows all permissions and bypasses channel permission overwrites.
    ADMINISTRATOR = 0x8
    #: Allows management and editing of channels.
    MANAGE_CHANNELS = 0x10
    #: Allows management and editing of the guild.
    MANAGE_GUILD = 0x20
    #: Allows for the addition of reactions to messages.
    ADD_REACTIONS = 0x40
    #: Allows for viewing of audit logs.
    VIEW_AUDIT_LOG = 0x80
    #: Allows for using priority speaker in a voice channel.
    PRIORITY_SPEAKER = 0x1_00
    #: Allows the user to go live.
    STREAM = 0x2_00
    #: Allows guild members to view a channel, which includes reading messages in text channels.
    VIEW_CHANNEL = 0x4_00
    #: Allows for sending messages in a channel.
    SEND_MESSAGES = 0x8_00
    #: Allows for sending of ``/tts`` messages.
    SEND_TTS_MESSAGES = 0x10_00
    #: Allows for deletion of other users messages.
    MANAGE_MESSAGES = 0x20_00
    #: Links sent by users with this permission will be auto-embedded.
    EMBED_LINKS = 0x40_00
    #: Allows for uploading images and files
    ATTACH_FILES = 0x80_00
    #: Allows for reading of message history.
    READ_MESSAGE_HISTORY = 0x1_00_00
    #: Allows for using the ``@everyone`` tag to notify all users in a channel, and the
    #: ``@here`` tag to notify all online users in a channel, and the ``@role`` tag (even
    #: if the role is not mentionable) to notify all users with that role in a channel.
    MENTION_EVERYONE = 0x2_00_00
    #: Allows the usage of custom emojis from other servers.
    USE_EXTERNAL_EMOJIS = 0x4_00_00
    #: Allows for joining of a voice channel.
    CONNECT = 0x10_00_00
    #: Allows for speaking in a voice channel.
    SPEAK = 0x20_00_00
    #: Allows for muting members in a voice channel.
    MUTE_MEMBERS = 0x40_00_00
    #: Allows for deafening of members in a voice channel.
    DEAFEN_MEMBERS = 0x80_00_00
    #: Allows for moving of members between voice channels.
    MOVE_MEMBERS = 0x1_00_00_00
    #: Allows for using voice-activity-detection in a voice channel.
    USE_VAD = 0x2_00_00_00
    #: Allows for modification of own nickname.
    CHANGE_NICKNAME = 0x4_00_00_00
    #: Allows for modification of other users nicknames.
    MANAGE_NICKNAMES = 0x8_00_00_00
    #: Allows management and editing of roles.
    MANAGE_ROLES = 0x10_00_00_00
    #: Allows management and editing of webhooks.
    MANAGE_WEBHOOKS = 0x20_00_00_00
    #: Allows management and editing of emojis.
    MANAGE_EMOJIS = 0x40_00_00_00



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe webhooks on Discord."""
__all__ = ["WebhookType", "Webhook"]

import enum
import typing

import attr

from hikari import entities
from hikari import snowflakes
from hikari import users
from hikari.internal import marshaller


@enum.unique
class WebhookType(enum.IntEnum):
    """Types of webhook."""

    #: Incoming webhook.
    INCOMING = 1
    #: Channel Follower webhook.
    CHANNEL_FOLLOWER = 2


@marshaller.marshallable()
@attr.s(slots=True)
class Webhook(snowflakes.UniqueEntity, entities.Deserializable):
    """Represents a webhook object on Discord.

    This is an endpoint that can have messages sent to it using standard
    HTTP requests, which enables external services that are not bots to
    send informational messages to specific channels.
    """

    #: The type of the webhook.
    #:
    #: :type: :obj:`WebhookType`
    type: WebhookType = marshaller.attrib(deserializer=WebhookType)

    #: The guild ID of the webhook.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`, optional
    guild_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )

    #: The channel ID this webhook is for.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`
    channel_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The user that created the webhook
    #:
    #: Note
    #: ----
    #: This will be ``None`` when getting a webhook with a token
    #:
    #:
    #: :type: :obj:`hikari.users.User`, optional
    user: typing.Optional[users.User] = marshaller.attrib(deserializer=users.User.deserialize, if_undefined=None)

    #: The default name of the webhook.
    #:
    #: :type: :obj:`str`, optional
    name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None)

    #: The default avatar hash of the webhook.
    #:
    #: :type: :obj:`str`, optional
    avatar_hash: typing.Optional[str] = marshaller.attrib(raw_name="avatar", deserializer=str, if_none=None)

    #: The token of the webhook.
    #:
    #: Note
    #: ----
    #: This is only available for Incoming webhooks.
    #:
    #:
    #: :type: :obj:`str`, optional
    token: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe both DMs and guild channels on Discord.

.. inheritance-diagram::
    hikari.channels
    enum.IntEnum
    hikari.entities.HikariEntity
    hikari.entities.Deserializable
    hikari.entities.Serializable
    hikari.snowflakes.UniqueEntity
    :parts: 1
"""

__all__ = [
    "Channel",
    "ChannelType",
    "PermissionOverwrite",
    "PermissionOverwriteType",
    "PartialChannel",
    "DMChannel",
    "GroupDMChannel",
    "GuildCategory",
    "GuildTextChannel",
    "GuildNewsChannel",
    "GuildStoreChannel",
    "GuildVoiceChannel",
]

import datetime
import enum
import typing

import attr

from hikari import entities
from hikari import permissions
from hikari import snowflakes
from hikari import users
from hikari.internal import marshaller
from hikari.internal import more_collections


@enum.unique
class ChannelType(enum.IntEnum):
    """The known channel types that are exposed to us by the api."""

    #: A text channel in a guild.
    GUILD_TEXT = 0
    #: A direct channel between two users.
    DM = 1
    #: A voice channel in a guild.
    GUILD_VOICE = 2
    #: A direct channel between multiple users.
    GROUP_DM = 3
    #: An category used for organizing channels in a guild.
    GUILD_CATEGORY = 4
    #: A channel that can be followed and can crosspost.
    GUILD_NEWS = 5
    #: A channel that show's a game's store page.
    GUILD_STORE = 6


@enum.unique
class PermissionOverwriteType(str, enum.Enum):
    """The type of entity a Permission Overwrite targets."""

    #: A permission overwrite that targets all the members with a specific
    #: guild role.
    ROLE = "role"
    #: A permission overwrite that targets a specific guild member.
    MEMBER = "member"


@marshaller.marshallable()
@attr.s(slots=True)
class PermissionOverwrite(snowflakes.UniqueEntity, entities.Deserializable, entities.Serializable):
    """Represents permission overwrites for a channel or role in a channel."""

    #: The type of entity this overwrite targets.
    #:
    #: :type: :obj:`PermissionOverwriteType`
    type: PermissionOverwriteType = marshaller.attrib(deserializer=PermissionOverwriteType)

    #: The permissions this overwrite allows.
    #:
    #: :type: :obj:`hikari.permissions.Permission`
    allow: permissions.Permission = marshaller.attrib(deserializer=permissions.Permission)

    #: The permissions this overwrite denies.
    #:
    #: :type: :obj:`hikari.permissions.Permission`
    deny: permissions.Permission = marshaller.attrib(deserializer=permissions.Permission)

    @property
    def unset(self) -> permissions.Permission:
        """Bitfield of all permissions not explicitly allowed or denied by this overwrite."""
        return typing.cast(permissions.Permission, (self.allow | self.deny))


def register_channel_type(type_: ChannelType) -> typing.Callable[[typing.Type["Channel"]], typing.Type["Channel"]]:
    """Generate a decorator for channel classes defined in this library.

    This allows them to associate themselves with a given channel type.

    Parameters
    ----------
    type_ : :obj:`ChannelType`
        The channel type to associate with.

    Returns
    -------
    ``decorator(T) -> T``
        The decorator to decorate the class with.
    """

    def decorator(cls):
        mapping = getattr(register_channel_type, "types", {})
        mapping[type_] = cls
        setattr(register_channel_type, "types", mapping)
        return cls

    return decorator


@marshaller.marshallable()
@attr.s(slots=True)
class Channel(snowflakes.UniqueEntity, entities.Deserializable):
    """Base class for all channels."""

    #: The channel's type.
    #:
    #: :type: :obj:`ChannelType`
    type: ChannelType = marshaller.attrib(deserializer=ChannelType)


@marshaller.marshallable()
@attr.s(slots=True)
class PartialChannel(Channel):
    """Represents a channel where we've only received it's basic information.

    This is commonly received in REST responses.
    """

    #: The channel's name.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)


@register_channel_type(ChannelType.DM)
@marshaller.marshallable()
@attr.s(slots=True)
class DMChannel(Channel):
    """Represents a DM channel."""

    #: The ID of the last message sent in this channel.
    #:
    #: Note
    #: ----
    #: This might point to an invalid or deleted message.
    #:
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`, optional
    last_message_id: snowflakes.Snowflake = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_none=None
    )

    #: The recipients of the DM.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`hikari.snowflakes.Snowflake`, :obj:`hikari.users.User` ]
    recipients: typing.Mapping[snowflakes.Snowflake, users.User] = marshaller.attrib(
        deserializer=lambda recipients: {user.id: user for user in map(users.User.deserialize, recipients)}
    )


@register_channel_type(ChannelType.GROUP_DM)
@marshaller.marshallable()
@attr.s(slots=True)
class GroupDMChannel(DMChannel):
    """Represents a DM group channel."""

    #: The group's name.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: The ID of the owner of the group.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`
    owner_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The hash of the icon of the group.
    #:
    #: :type: :obj:`str`, optional
    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str, if_none=None)

    #: The ID of the application that created the group DM, if it's a
    #: bot based group DM.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`, optional
    application_id: typing.Optional[snowflakes.Snowflake] = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_undefined=None
    )


@marshaller.marshallable()
@attr.s(slots=True)
class GuildChannel(Channel):
    """The base for anything that is a guild channel."""

    #: The ID of the guild the channel belongs to.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`
    guild_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize)

    #: The sorting position of the channel.
    #:
    #: :type: :obj:`int`
    position: int = marshaller.attrib(deserializer=int)

    #: The permission overwrites for the channel.
    #:
    #: :type: :obj:`typing.Mapping` [ :obj:`hikari.snowflakes.Snowflake`, :obj:`PermissionOverwrite` ]
    permission_overwrites: PermissionOverwrite = marshaller.attrib(
        deserializer=lambda overwrites: {o.id: o for o in map(PermissionOverwrite.deserialize, overwrites)}
    )

    #: The name of the channel.
    #:
    #: :type: :obj:`str`
    name: str = marshaller.attrib(deserializer=str)

    #: Wheter the channel is marked as NSFW.
    #:
    #: :type: :obj:`bool`
    is_nsfw: bool = marshaller.attrib(raw_name="nsfw", deserializer=bool)

    #: The ID of the parent category the channel belongs to.
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`, optional
    parent_id: snowflakes.Snowflake = marshaller.attrib(deserializer=snowflakes.Snowflake.deserialize, if_none=None)


@register_channel_type(ChannelType.GUILD_CATEGORY)
@marshaller.marshallable()
@attr.s(slots=True)
class GuildCategory(GuildChannel):
    """Represents a guild category."""


@register_channel_type(ChannelType.GUILD_TEXT)
@marshaller.marshallable()
@attr.s(slots=True)
class GuildTextChannel(GuildChannel):
    """Represents a guild text channel."""

    #: The topic of the channel.
    #:
    #: :type: :obj:`str`, optional
    topic: str = marshaller.attrib(deserializer=str, if_none=None)

    #: The ID of the last message sent in this channel.
    #:
    #: Note
    #: ----
    #: This might point to an invalid or deleted message.
    #:
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`, optional
    last_message_id: snowflakes.Snowflake = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_none=None
    )

    #: The delay (in seconds) between a user can send a message
    #: to this channel.
    #:
    #: Note
    #: ----
    #: Bots, as well as users with ``MANAGE_MESSAGES`` or
    #: ``MANAGE_CHANNEL``, are not afected by this.
    #:
    #: :type: :obj:`datetime.timedelta`
    rate_limit_per_user: datetime.timedelta = marshaller.attrib(
        deserializer=lambda payload: datetime.timedelta(seconds=payload)
    )


@register_channel_type(ChannelType.GUILD_NEWS)
@marshaller.marshallable()
@attr.s(slots=True)
class GuildNewsChannel(GuildChannel):
    """Represents an news channel."""

    #: The topic of the channel.
    #:
    #: :type: :obj:`str`, optional
    topic: str = marshaller.attrib(deserializer=str, if_none=None)

    #: The ID of the last message sent in this channel.
    #:
    #: Note
    #: ----
    #: This might point to an invalid or deleted message.
    #:
    #:
    #: :type: :obj:`hikari.snowflakes.Snowflake`, optional
    last_message_id: snowflakes.Snowflake = marshaller.attrib(
        deserializer=snowflakes.Snowflake.deserialize, if_none=None
    )


@register_channel_type(ChannelType.GUILD_STORE)
@marshaller.marshallable()
@attr.s(slots=True)
class GuildStoreChannel(GuildChannel):
    """Represents a store channel."""


@register_channel_type(ChannelType.GUILD_VOICE)
@marshaller.marshallable()
@attr.s(slots=True)
class GuildVoiceChannel(GuildChannel):
    """Represents an voice channel."""

    #: The bitrate for the voice channel (in bits).
    #:
    #: :type: :obj:`int`
    bitrate: int = marshaller.attrib(deserializer=int)

    #: The user limit for the voice channel.
    #:
    #: :type: :obj:`int`
    user_limit: int = marshaller.attrib(deserializer=int)


def deserialize_channel(payload: typing.Dict[str, typing.Any]) -> typing.Union[GuildChannel, DMChannel]:
    """Deserialize a channel object into the corresponding class.

    Warning
    -------
    This can only be used to deserialize full channel objects. To deserialize a
    partial object, use ``PartialChannel.deserialize()``.
    """
    type_id = payload["type"]
    types = getattr(register_channel_type, "types", more_collections.EMPTY_DICT)
    channel_type = types[type_id]
    return channel_type.deserialize(payload)
